/* header auto-generated by pidl */

#ifndef _PIDL_HEADER_server_id
#define _PIDL_HEADER_server_id

#include <stdint.h>

#include <core/ntstatus.h>

#ifndef _HEADER_server_id
#define _HEADER_server_id

#define NONCLUSTER_VNN	( 0xFFFFFFFF )
#define SERVERID_UNIQUE_ID_NOT_TO_VERIFY	( 0xFFFFFFFFFFFFFFFFULL )
struct server_id {
	uint64_t pid;
	uint32_t task_id;
	uint32_t vnn;
	uint64_t unique_id;
}/* [public] */;

#endif /* _HEADER_server_id */
#endif /* _PIDL_HEADER_server_id */
