/* header auto-generated by pidl */

#include <ndr.h>
#include <gen_ndr/svcctl.h>

#ifndef _HEADER_NDR_svcctl
#define _HEADER_NDR_svcctl

#include <ndr/ndr_svcctl.h>
#define NDR_SVCCTL_UUID "367abb81-9844-35f1-ad32-98f038001003"
#define NDR_SVCCTL_VERSION 2.0
#define NDR_SVCCTL_NAME "svcctl"
#define NDR_SVCCTL_HELPSTRING "Service Control"
extern const struct ndr_interface_table ndr_table_svcctl;
#define NDR_SVCCTL_CLOSESERVICEHANDLE (0x00)

#define NDR_SVCCTL_CONTROLSERVICE (0x01)

#define NDR_SVCCTL_DELETESERVICE (0x02)

#define NDR_SVCCTL_LOCKSERVICEDATABASE (0x03)

#define NDR_SVCCTL_QUERYSERVICEOBJECTSECURITY (0x04)

#define NDR_SVCCTL_SETSERVICEOBJECTSECURITY (0x05)

#define NDR_SVCCTL_QUERYSERVICESTATUS (0x06)

#define NDR_SVCCTL_SETSERVICESTATUS (0x07)

#define NDR_SVCCTL_UNLOCKSERVICEDATABASE (0x08)

#define NDR_SVCCTL_NOTIFYBOOTCONFIGSTATUS (0x09)

#define NDR_SVCCTL_SCSETSERVICEBITSW (0x0a)

#define NDR_SVCCTL_CHANGESERVICECONFIGW (0x0b)

#define NDR_SVCCTL_CREATESERVICEW (0x0c)

#define NDR_SVCCTL_ENUMDEPENDENTSERVICESW (0x0d)

#define NDR_SVCCTL_ENUMSERVICESSTATUSW (0x0e)

#define NDR_SVCCTL_OPENSCMANAGERW (0x0f)

#define NDR_SVCCTL_OPENSERVICEW (0x10)

#define NDR_SVCCTL_QUERYSERVICECONFIGW (0x11)

#define NDR_SVCCTL_QUERYSERVICELOCKSTATUSW (0x12)

#define NDR_SVCCTL_STARTSERVICEW (0x13)

#define NDR_SVCCTL_GETSERVICEDISPLAYNAMEW (0x14)

#define NDR_SVCCTL_GETSERVICEKEYNAMEW (0x15)

#define NDR_SVCCTL_SCSETSERVICEBITSA (0x16)

#define NDR_SVCCTL_CHANGESERVICECONFIGA (0x17)

#define NDR_SVCCTL_CREATESERVICEA (0x18)

#define NDR_SVCCTL_ENUMDEPENDENTSERVICESA (0x19)

#define NDR_SVCCTL_ENUMSERVICESSTATUSA (0x1a)

#define NDR_SVCCTL_OPENSCMANAGERA (0x1b)

#define NDR_SVCCTL_OPENSERVICEA (0x1c)

#define NDR_SVCCTL_QUERYSERVICECONFIGA (0x1d)

#define NDR_SVCCTL_QUERYSERVICELOCKSTATUSA (0x1e)

#define NDR_SVCCTL_STARTSERVICEA (0x1f)

#define NDR_SVCCTL_GETSERVICEDISPLAYNAMEA (0x20)

#define NDR_SVCCTL_GETSERVICEKEYNAMEA (0x21)

#define NDR_SVCCTL_GETCURRENTGROUPESTATEW (0x22)

#define NDR_SVCCTL_ENUMSERVICEGROUPW (0x23)

#define NDR_SVCCTL_CHANGESERVICECONFIG2A (0x24)

#define NDR_SVCCTL_CHANGESERVICECONFIG2W (0x25)

#define NDR_SVCCTL_QUERYSERVICECONFIG2A (0x26)

#define NDR_SVCCTL_QUERYSERVICECONFIG2W (0x27)

#define NDR_SVCCTL_QUERYSERVICESTATUSEX (0x28)

#define NDR_SVCCTL_ENUMSERVICESSTATUSEXA (0x29)

#define NDR_SVCCTL_ENUMSERVICESSTATUSEXW (0x2a)

#define NDR_SVCCTL_SCSENDTSMESSAGE (0x2b)

#define NDR_SVCCTL_CREATESERVICEWOW64A (0x2c)

#define NDR_SVCCTL_CREATESERVICEWOW64W (0x2d)

#define NDR_OPNUM46NOTUSEDONWIRE (0x2e)

#define NDR_SVCCTL_NOTIFYSERVICESTATUSCHANGE (0x2f)

#define NDR_SVCCTL_GETNOTIFYRESULTS (0x30)

#define NDR_SVCCTL_CLOSENOTIFYHANDLE (0x31)

#define NDR_SVCCTL_CONTROLSERVICEEXA (0x32)

#define NDR_SVCCTL_CONTROLSERVICEEXW (0x33)

#define NDR_OPNUM52NOTUSEDONWIRE (0x34)

#define NDR_OPNUM53NOTUSEDONWIRE (0x35)

#define NDR_OPNUM54NOTUSEDONWIRE (0x36)

#define NDR_OPNUM55NOTUSEDONWIRE (0x37)

#define NDR_SVCCTL_QUERYSERVICECONFIGEX (0x38)

#define NDR_OPNUM57NOTUSEDONWIRE (0x39)

#define NDR_OPNUM58NOTUSEDONWIRE (0x3a)

#define NDR_OPNUM59NOTUSEDONWIRE (0x3b)

#define NDR_SVCCTL_CREATEWOWSERVICE (0x3c)

#define NDR_SVCCTL_OPENSCMANAGER2 (0x3d)

#define NDR_SVCCTL_CALL_COUNT (62)
void ndr_print_SERVICE_LOCK_STATUS(struct ndr_print *ndr, const char *name, const struct SERVICE_LOCK_STATUS *r);
void ndr_print_svcctl_ServiceStatus(struct ndr_print *ndr, const char *name, enum svcctl_ServiceStatus r);
void ndr_print_svcctl_ControlsAccepted(struct ndr_print *ndr, const char *name, uint32_t r);
void ndr_print_SERVICE_STATUS(struct ndr_print *ndr, const char *name, const struct SERVICE_STATUS *r);
enum ndr_err_code ndr_push_SERVICE_STATUS_PROCESS(struct ndr_push *ndr, ndr_flags_type ndr_flags, const struct SERVICE_STATUS_PROCESS *r);
enum ndr_err_code ndr_pull_SERVICE_STATUS_PROCESS(struct ndr_pull *ndr, ndr_flags_type ndr_flags, struct SERVICE_STATUS_PROCESS *r);
void ndr_print_SERVICE_STATUS_PROCESS(struct ndr_print *ndr, const char *name, const struct SERVICE_STATUS_PROCESS *r);
enum ndr_err_code ndr_push_ENUM_SERVICE_STATUSW(struct ndr_push *ndr, ndr_flags_type ndr_flags, const struct ENUM_SERVICE_STATUSW *r);
enum ndr_err_code ndr_pull_ENUM_SERVICE_STATUSW(struct ndr_pull *ndr, ndr_flags_type ndr_flags, struct ENUM_SERVICE_STATUSW *r);
void ndr_print_ENUM_SERVICE_STATUSW(struct ndr_print *ndr, const char *name, const struct ENUM_SERVICE_STATUSW *r);
size_t ndr_size_ENUM_SERVICE_STATUSW(const struct ENUM_SERVICE_STATUSW *r, libndr_flags flags);
enum ndr_err_code ndr_push_ENUM_SERVICE_STATUSA(struct ndr_push *ndr, ndr_flags_type ndr_flags, const struct ENUM_SERVICE_STATUSA *r);
enum ndr_err_code ndr_pull_ENUM_SERVICE_STATUSA(struct ndr_pull *ndr, ndr_flags_type ndr_flags, struct ENUM_SERVICE_STATUSA *r);
void ndr_print_ENUM_SERVICE_STATUSA(struct ndr_print *ndr, const char *name, const struct ENUM_SERVICE_STATUSA *r);
size_t ndr_size_ENUM_SERVICE_STATUSA(const struct ENUM_SERVICE_STATUSA *r, libndr_flags flags);
void ndr_print_SERVICE_CONTROL(struct ndr_print *ndr, const char *name, enum SERVICE_CONTROL r);
void ndr_print_svcctl_ErrorControl(struct ndr_print *ndr, const char *name, enum svcctl_ErrorControl r);
void ndr_print_svcctl_StartType(struct ndr_print *ndr, const char *name, enum svcctl_StartType r);
void ndr_print_svcctl_ServiceState(struct ndr_print *ndr, const char *name, enum svcctl_ServiceState r);
void ndr_print_svcctl_MgrAccessMask(struct ndr_print *ndr, const char *name, uint32_t r);
void ndr_print_svcctl_ServiceAccessMask(struct ndr_print *ndr, const char *name, uint32_t r);
enum ndr_err_code ndr_push_QUERY_SERVICE_CONFIG(struct ndr_push *ndr, ndr_flags_type ndr_flags, const struct QUERY_SERVICE_CONFIG *r);
enum ndr_err_code ndr_pull_QUERY_SERVICE_CONFIG(struct ndr_pull *ndr, ndr_flags_type ndr_flags, struct QUERY_SERVICE_CONFIG *r);
void ndr_print_QUERY_SERVICE_CONFIG(struct ndr_print *ndr, const char *name, const struct QUERY_SERVICE_CONFIG *r);
size_t ndr_size_QUERY_SERVICE_CONFIG(const struct QUERY_SERVICE_CONFIG *r, libndr_flags flags);
void ndr_print_svcctl_ArgumentString(struct ndr_print *ndr, const char *name, const struct svcctl_ArgumentString *r);
void ndr_print_svcctl_ConfigLevel(struct ndr_print *ndr, const char *name, enum svcctl_ConfigLevel r);
enum ndr_err_code ndr_push_SERVICE_DESCRIPTION(struct ndr_push *ndr, ndr_flags_type ndr_flags, const struct SERVICE_DESCRIPTION *r);
enum ndr_err_code ndr_pull_SERVICE_DESCRIPTION(struct ndr_pull *ndr, ndr_flags_type ndr_flags, struct SERVICE_DESCRIPTION *r);
void ndr_print_SERVICE_DESCRIPTION(struct ndr_print *ndr, const char *name, const struct SERVICE_DESCRIPTION *r);
size_t ndr_size_SERVICE_DESCRIPTION(const struct SERVICE_DESCRIPTION *r, libndr_flags flags);
void ndr_print_SC_ACTION_TYPE(struct ndr_print *ndr, const char *name, enum SC_ACTION_TYPE r);
void ndr_print_SC_ACTION(struct ndr_print *ndr, const char *name, const struct SC_ACTION *r);
enum ndr_err_code ndr_push_SERVICE_FAILURE_ACTIONSW(struct ndr_push *ndr, ndr_flags_type ndr_flags, const struct SERVICE_FAILURE_ACTIONSW *r);
enum ndr_err_code ndr_pull_SERVICE_FAILURE_ACTIONSW(struct ndr_pull *ndr, ndr_flags_type ndr_flags, struct SERVICE_FAILURE_ACTIONSW *r);
void ndr_print_SERVICE_FAILURE_ACTIONSW(struct ndr_print *ndr, const char *name, const struct SERVICE_FAILURE_ACTIONSW *r);
size_t ndr_size_SERVICE_FAILURE_ACTIONSW(const struct SERVICE_FAILURE_ACTIONSW *r, libndr_flags flags);
void ndr_print_svcctl_StatusLevel(struct ndr_print *ndr, const char *name, enum svcctl_StatusLevel r);
void ndr_print_SERVICE_NOTIFY_STATUS_CHANGE_PARAMS_1(struct ndr_print *ndr, const char *name, const struct SERVICE_NOTIFY_STATUS_CHANGE_PARAMS_1 *r);
void ndr_print_SERVICE_NOTIFY_STATUS_CHANGE_PARAMS_2(struct ndr_print *ndr, const char *name, const struct SERVICE_NOTIFY_STATUS_CHANGE_PARAMS_2 *r);
void ndr_print_SC_RPC_NOTIFY_PARAMS_u(struct ndr_print *ndr, const char *name, const union SC_RPC_NOTIFY_PARAMS_u *r);
void ndr_print_SC_RPC_NOTIFY_PARAMS(struct ndr_print *ndr, const char *name, const struct SC_RPC_NOTIFY_PARAMS *r);
void ndr_print_SC_RPC_NOTIFY_PARAMS_LIST(struct ndr_print *ndr, const char *name, const struct SC_RPC_NOTIFY_PARAMS_LIST *r);
void ndr_print_SERVICE_CONTROL_STATUS_REASON_IN_PARAMSA(struct ndr_print *ndr, const char *name, const struct SERVICE_CONTROL_STATUS_REASON_IN_PARAMSA *r);
void ndr_print_SERVICE_CONTROL_STATUS_REASON_OUT_PARAMS(struct ndr_print *ndr, const char *name, const struct SERVICE_CONTROL_STATUS_REASON_OUT_PARAMS *r);
void ndr_print_SC_RPC_SERVICE_CONTROL_IN_PARAMSA(struct ndr_print *ndr, const char *name, const union SC_RPC_SERVICE_CONTROL_IN_PARAMSA *r);
void ndr_print_SC_RPC_SERVICE_CONTROL_OUT_PARAMSA(struct ndr_print *ndr, const char *name, const union SC_RPC_SERVICE_CONTROL_OUT_PARAMSA *r);
void ndr_print_SERVICE_CONTROL_STATUS_REASON_IN_PARAMSW(struct ndr_print *ndr, const char *name, const struct SERVICE_CONTROL_STATUS_REASON_IN_PARAMSW *r);
void ndr_print_SC_RPC_SERVICE_CONTROL_IN_PARAMSW(struct ndr_print *ndr, const char *name, const union SC_RPC_SERVICE_CONTROL_IN_PARAMSW *r);
void ndr_print_SC_RPC_SERVICE_CONTROL_OUT_PARAMSW(struct ndr_print *ndr, const char *name, const union SC_RPC_SERVICE_CONTROL_OUT_PARAMSW *r);
void ndr_print_SERVICE_DESCRIPTIONW(struct ndr_print *ndr, const char *name, const struct SERVICE_DESCRIPTIONW *r);
void ndr_print_SERVICE_DELAYED_AUTO_START_INFO(struct ndr_print *ndr, const char *name, const struct SERVICE_DELAYED_AUTO_START_INFO *r);
void ndr_print_SERVICE_FAILURE_ACTIONS_FLAG(struct ndr_print *ndr, const char *name, const struct SERVICE_FAILURE_ACTIONS_FLAG *r);
void ndr_print_SERVICE_SID_INFO(struct ndr_print *ndr, const char *name, const struct SERVICE_SID_INFO *r);
void ndr_print_SERVICE_RPC_REQUIRED_PRIVILEGES_INFO(struct ndr_print *ndr, const char *name, const struct SERVICE_RPC_REQUIRED_PRIVILEGES_INFO *r);
void ndr_print_SERVICE_PRESHUTDOWN_INFO(struct ndr_print *ndr, const char *name, const struct SERVICE_PRESHUTDOWN_INFO *r);
void ndr_print_SERVICE_TRIGGER_SPECIFIC_DATA_ITEM(struct ndr_print *ndr, const char *name, const struct SERVICE_TRIGGER_SPECIFIC_DATA_ITEM *r);
void ndr_print_SERVICE_TRIGGER(struct ndr_print *ndr, const char *name, const struct SERVICE_TRIGGER *r);
void ndr_print_SERVICE_TRIGGER_INFO(struct ndr_print *ndr, const char *name, const struct SERVICE_TRIGGER_INFO *r);
void ndr_print_SERVICE_PREFERRED_NODE_INFO(struct ndr_print *ndr, const char *name, const struct SERVICE_PREFERRED_NODE_INFO *r);
void ndr_print_SC_RPC_CONFIG_INFOW_u(struct ndr_print *ndr, const char *name, const union SC_RPC_CONFIG_INFOW_u *r);
void ndr_print_SC_RPC_CONFIG_INFOW(struct ndr_print *ndr, const char *name, const struct SC_RPC_CONFIG_INFOW *r);
void ndr_print_svcctl_CloseServiceHandle(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_CloseServiceHandle *r);
void ndr_print_svcctl_ControlService(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_ControlService *r);
void ndr_print_svcctl_DeleteService(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_DeleteService *r);
void ndr_print_svcctl_LockServiceDatabase(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_LockServiceDatabase *r);
void ndr_print_svcctl_QueryServiceObjectSecurity(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_QueryServiceObjectSecurity *r);
void ndr_print_svcctl_SetServiceObjectSecurity(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_SetServiceObjectSecurity *r);
void ndr_print_svcctl_QueryServiceStatus(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_QueryServiceStatus *r);
void ndr_print_svcctl_SetServiceStatus(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_SetServiceStatus *r);
void ndr_print_svcctl_UnlockServiceDatabase(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_UnlockServiceDatabase *r);
void ndr_print_svcctl_NotifyBootConfigStatus(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_NotifyBootConfigStatus *r);
void ndr_print_svcctl_SCSetServiceBitsW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_SCSetServiceBitsW *r);
enum ndr_err_code ndr_push_svcctl_ChangeServiceConfigW(struct ndr_push *ndr, ndr_flags_type flags, const struct svcctl_ChangeServiceConfigW *r);
enum ndr_err_code ndr_pull_svcctl_ChangeServiceConfigW(struct ndr_pull *ndr, ndr_flags_type flags, struct svcctl_ChangeServiceConfigW *r);
void ndr_print_svcctl_ChangeServiceConfigW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_ChangeServiceConfigW *r);
void ndr_print_svcctl_CreateServiceW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_CreateServiceW *r);
void ndr_print_svcctl_EnumDependentServicesW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_EnumDependentServicesW *r);
void ndr_print_svcctl_EnumServicesStatusW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_EnumServicesStatusW *r);
void ndr_print_svcctl_OpenSCManagerW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_OpenSCManagerW *r);
void ndr_print_svcctl_OpenServiceW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_OpenServiceW *r);
void ndr_print_svcctl_QueryServiceConfigW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_QueryServiceConfigW *r);
void ndr_print_svcctl_QueryServiceLockStatusW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_QueryServiceLockStatusW *r);
void ndr_print_svcctl_StartServiceW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_StartServiceW *r);
void ndr_print_svcctl_GetServiceDisplayNameW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_GetServiceDisplayNameW *r);
void ndr_print_svcctl_GetServiceKeyNameW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_GetServiceKeyNameW *r);
void ndr_print_svcctl_SCSetServiceBitsA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_SCSetServiceBitsA *r);
void ndr_print_svcctl_ChangeServiceConfigA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_ChangeServiceConfigA *r);
void ndr_print_svcctl_CreateServiceA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_CreateServiceA *r);
void ndr_print_svcctl_EnumDependentServicesA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_EnumDependentServicesA *r);
void ndr_print_svcctl_EnumServicesStatusA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_EnumServicesStatusA *r);
void ndr_print_svcctl_OpenSCManagerA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_OpenSCManagerA *r);
void ndr_print_svcctl_OpenServiceA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_OpenServiceA *r);
void ndr_print_svcctl_QueryServiceConfigA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_QueryServiceConfigA *r);
void ndr_print_svcctl_QueryServiceLockStatusA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_QueryServiceLockStatusA *r);
void ndr_print_svcctl_StartServiceA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_StartServiceA *r);
void ndr_print_svcctl_GetServiceDisplayNameA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_GetServiceDisplayNameA *r);
void ndr_print_svcctl_GetServiceKeyNameA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_GetServiceKeyNameA *r);
void ndr_print_svcctl_GetCurrentGroupeStateW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_GetCurrentGroupeStateW *r);
void ndr_print_svcctl_EnumServiceGroupW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_EnumServiceGroupW *r);
void ndr_print_svcctl_ChangeServiceConfig2A(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_ChangeServiceConfig2A *r);
void ndr_print_svcctl_ChangeServiceConfig2W(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_ChangeServiceConfig2W *r);
void ndr_print_svcctl_QueryServiceConfig2A(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_QueryServiceConfig2A *r);
void ndr_print_svcctl_QueryServiceConfig2W(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_QueryServiceConfig2W *r);
void ndr_print_svcctl_QueryServiceStatusEx(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_QueryServiceStatusEx *r);
void ndr_print_svcctl_EnumServicesStatusExA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_EnumServicesStatusExA *r);
void ndr_print_svcctl_EnumServicesStatusExW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_EnumServicesStatusExW *r);
void ndr_print_svcctl_SCSendTSMessage(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_SCSendTSMessage *r);
void ndr_print_svcctl_CreateServiceWOW64A(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_CreateServiceWOW64A *r);
void ndr_print_svcctl_CreateServiceWOW64W(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_CreateServiceWOW64W *r);
void ndr_print_Opnum46NotUsedOnWire(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct Opnum46NotUsedOnWire *r);
void ndr_print_svcctl_NotifyServiceStatusChange(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_NotifyServiceStatusChange *r);
void ndr_print_svcctl_GetNotifyResults(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_GetNotifyResults *r);
void ndr_print_svcctl_CloseNotifyHandle(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_CloseNotifyHandle *r);
void ndr_print_svcctl_ControlServiceExA(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_ControlServiceExA *r);
void ndr_print_svcctl_ControlServiceExW(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_ControlServiceExW *r);
void ndr_print_Opnum52NotUsedOnWire(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct Opnum52NotUsedOnWire *r);
void ndr_print_Opnum53NotUsedOnWire(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct Opnum53NotUsedOnWire *r);
void ndr_print_Opnum54NotUsedOnWire(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct Opnum54NotUsedOnWire *r);
void ndr_print_Opnum55NotUsedOnWire(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct Opnum55NotUsedOnWire *r);
void ndr_print_svcctl_QueryServiceConfigEx(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_QueryServiceConfigEx *r);
void ndr_print_Opnum57NotUsedOnWire(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct Opnum57NotUsedOnWire *r);
void ndr_print_Opnum58NotUsedOnWire(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct Opnum58NotUsedOnWire *r);
void ndr_print_Opnum59NotUsedOnWire(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct Opnum59NotUsedOnWire *r);
void ndr_print_svcctl_CreateWowService(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_CreateWowService *r);
void ndr_print_svcctl_OpenSCManager2(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct svcctl_OpenSCManager2 *r);
#endif /* _HEADER_NDR_svcctl */
