/* header auto-generated by pidl */

#include <ndr.h>
#include <gen_ndr/samr.h>

#ifndef _HEADER_NDR_samr
#define _HEADER_NDR_samr

#define NDR_SAMR_UUID "12345778-1234-abcd-ef00-0123456789ac"
#define NDR_SAMR_VERSION 1.0
#define NDR_SAMR_NAME "samr"
#define NDR_SAMR_HELPSTRING NULL
extern const struct ndr_interface_table ndr_table_samr;
#define NDR_SAMR_CONNECT (0x00)

#define NDR_SAMR_CLOSE (0x01)

#define NDR_SAMR_SETSECURITY (0x02)

#define NDR_SAMR_QUERYSECURITY (0x03)

#define NDR_SAMR_SHUTDOWN (0x04)

#define NDR_SAMR_LOOKUPDOMAIN (0x05)

#define NDR_SAMR_ENUMDOMAINS (0x06)

#define NDR_SAMR_OPENDOMAIN (0x07)

#define NDR_SAMR_QUERYDOMAININFO (0x08)

#define NDR_SAMR_SETDOMAININFO (0x09)

#define NDR_SAMR_CREATEDOMAINGROUP (0x0a)

#define NDR_SAMR_ENUMDOMAINGROUPS (0x0b)

#define NDR_SAMR_CREATEUSER (0x0c)

#define NDR_SAMR_ENUMDOMAINUSERS (0x0d)

#define NDR_SAMR_CREATEDOMALIAS (0x0e)

#define NDR_SAMR_ENUMDOMAINALIASES (0x0f)

#define NDR_SAMR_GETALIASMEMBERSHIP (0x10)

#define NDR_SAMR_LOOKUPNAMES (0x11)

#define NDR_SAMR_LOOKUPRIDS (0x12)

#define NDR_SAMR_OPENGROUP (0x13)

#define NDR_SAMR_QUERYGROUPINFO (0x14)

#define NDR_SAMR_SETGROUPINFO (0x15)

#define NDR_SAMR_ADDGROUPMEMBER (0x16)

#define NDR_SAMR_DELETEDOMAINGROUP (0x17)

#define NDR_SAMR_DELETEGROUPMEMBER (0x18)

#define NDR_SAMR_QUERYGROUPMEMBER (0x19)

#define NDR_SAMR_SETMEMBERATTRIBUTESOFGROUP (0x1a)

#define NDR_SAMR_OPENALIAS (0x1b)

#define NDR_SAMR_QUERYALIASINFO (0x1c)

#define NDR_SAMR_SETALIASINFO (0x1d)

#define NDR_SAMR_DELETEDOMALIAS (0x1e)

#define NDR_SAMR_ADDALIASMEMBER (0x1f)

#define NDR_SAMR_DELETEALIASMEMBER (0x20)

#define NDR_SAMR_GETMEMBERSINALIAS (0x21)

#define NDR_SAMR_OPENUSER (0x22)

#define NDR_SAMR_DELETEUSER (0x23)

#define NDR_SAMR_QUERYUSERINFO (0x24)

#define NDR_SAMR_SETUSERINFO (0x25)

#define NDR_SAMR_CHANGEPASSWORDUSER (0x26)

#define NDR_SAMR_GETGROUPSFORUSER (0x27)

#define NDR_SAMR_QUERYDISPLAYINFO (0x28)

#define NDR_SAMR_GETDISPLAYENUMERATIONINDEX (0x29)

#define NDR_SAMR_TESTPRIVATEFUNCTIONSDOMAIN (0x2a)

#define NDR_SAMR_TESTPRIVATEFUNCTIONSUSER (0x2b)

#define NDR_SAMR_GETUSERPWINFO (0x2c)

#define NDR_SAMR_REMOVEMEMBERFROMFOREIGNDOMAIN (0x2d)

#define NDR_SAMR_QUERYDOMAININFO2 (0x2e)

#define NDR_SAMR_QUERYUSERINFO2 (0x2f)

#define NDR_SAMR_QUERYDISPLAYINFO2 (0x30)

#define NDR_SAMR_GETDISPLAYENUMERATIONINDEX2 (0x31)

#define NDR_SAMR_CREATEUSER2 (0x32)

#define NDR_SAMR_QUERYDISPLAYINFO3 (0x33)

#define NDR_SAMR_ADDMULTIPLEMEMBERSTOALIAS (0x34)

#define NDR_SAMR_REMOVEMULTIPLEMEMBERSFROMALIAS (0x35)

#define NDR_SAMR_OEMCHANGEPASSWORDUSER2 (0x36)

#define NDR_SAMR_CHANGEPASSWORDUSER2 (0x37)

#define NDR_SAMR_GETDOMPWINFO (0x38)

#define NDR_SAMR_CONNECT2 (0x39)

#define NDR_SAMR_SETUSERINFO2 (0x3a)

#define NDR_SAMR_SETBOOTKEYINFORMATION (0x3b)

#define NDR_SAMR_GETBOOTKEYINFORMATION (0x3c)

#define NDR_SAMR_CONNECT3 (0x3d)

#define NDR_SAMR_CONNECT4 (0x3e)

#define NDR_SAMR_CHANGEPASSWORDUSER3 (0x3f)

#define NDR_SAMR_CONNECT5 (0x40)

#define NDR_SAMR_RIDTOSID (0x41)

#define NDR_SAMR_SETDSRMPASSWORD (0x42)

#define NDR_SAMR_VALIDATEPASSWORD (0x43)

#define NDR_SAMR_OPNUM68NOTUSEDONWIRE (0x44)

#define NDR_SAMR_OPNUM69NOTUSEDONWIRE (0x45)

#define NDR_SAMR_OPNUM70NOTUSEDONWIRE (0x46)

#define NDR_SAMR_OPNUM71NOTUSEDONWIRE (0x47)

#define NDR_SAMR_OPNUM72NOTUSEDONWIRE (0x48)

#define NDR_SAMR_CHANGEPASSWORDUSER4 (0x49)

#define NDR_SAMR_CALL_COUNT (74)
enum ndr_err_code ndr_push_samr_AcctFlags(struct ndr_push *ndr, ndr_flags_type ndr_flags, uint32_t r);
enum ndr_err_code ndr_pull_samr_AcctFlags(struct ndr_pull *ndr, ndr_flags_type ndr_flags, uint32_t *r);
void ndr_print_samr_AcctFlags(struct ndr_print *ndr, const char *name, uint32_t r);
void ndr_print_samr_ConnectAccessMask(struct ndr_print *ndr, const char *name, uint32_t r);
void ndr_print_samr_UserAccessMask(struct ndr_print *ndr, const char *name, uint32_t r);
void ndr_print_samr_DomainAccessMask(struct ndr_print *ndr, const char *name, uint32_t r);
void ndr_print_samr_GroupAccessMask(struct ndr_print *ndr, const char *name, uint32_t r);
void ndr_print_samr_AliasAccessMask(struct ndr_print *ndr, const char *name, uint32_t r);
void ndr_print_samr_SamEntry(struct ndr_print *ndr, const char *name, const struct samr_SamEntry *r);
void ndr_print_samr_SamArray(struct ndr_print *ndr, const char *name, const struct samr_SamArray *r);
void ndr_print_samr_DomainInfoClass(struct ndr_print *ndr, const char *name, enum samr_DomainInfoClass r);
void ndr_print_samr_Role(struct ndr_print *ndr, const char *name, enum samr_Role r);
enum ndr_err_code ndr_push_samr_PasswordProperties(struct ndr_push *ndr, ndr_flags_type ndr_flags, uint32_t r);
enum ndr_err_code ndr_pull_samr_PasswordProperties(struct ndr_pull *ndr, ndr_flags_type ndr_flags, uint32_t *r);
void ndr_print_samr_PasswordProperties(struct ndr_print *ndr, const char *name, uint32_t r);
void ndr_print_samr_DomainServerState(struct ndr_print *ndr, const char *name, enum samr_DomainServerState r);
enum ndr_err_code ndr_push_samr_DomInfo1(struct ndr_push *ndr, ndr_flags_type ndr_flags, const struct samr_DomInfo1 *r);
enum ndr_err_code ndr_pull_samr_DomInfo1(struct ndr_pull *ndr, ndr_flags_type ndr_flags, struct samr_DomInfo1 *r);
void ndr_print_samr_DomInfo1(struct ndr_print *ndr, const char *name, const struct samr_DomInfo1 *r);
void ndr_print_samr_DomGeneralInformation(struct ndr_print *ndr, const char *name, const struct samr_DomGeneralInformation *r);
void ndr_print_samr_DomInfo3(struct ndr_print *ndr, const char *name, const struct samr_DomInfo3 *r);
void ndr_print_samr_DomOEMInformation(struct ndr_print *ndr, const char *name, const struct samr_DomOEMInformation *r);
void ndr_print_samr_DomInfo5(struct ndr_print *ndr, const char *name, const struct samr_DomInfo5 *r);
void ndr_print_samr_DomInfo6(struct ndr_print *ndr, const char *name, const struct samr_DomInfo6 *r);
void ndr_print_samr_DomInfo7(struct ndr_print *ndr, const char *name, const struct samr_DomInfo7 *r);
void ndr_print_samr_DomInfo8(struct ndr_print *ndr, const char *name, const struct samr_DomInfo8 *r);
void ndr_print_samr_DomInfo9(struct ndr_print *ndr, const char *name, const struct samr_DomInfo9 *r);
void ndr_print_samr_DomGeneralInformation2(struct ndr_print *ndr, const char *name, const struct samr_DomGeneralInformation2 *r);
void ndr_print_samr_DomInfo12(struct ndr_print *ndr, const char *name, const struct samr_DomInfo12 *r);
void ndr_print_samr_DomInfo13(struct ndr_print *ndr, const char *name, const struct samr_DomInfo13 *r);
void ndr_print_samr_DomainInfo(struct ndr_print *ndr, const char *name, const union samr_DomainInfo *r);
void ndr_print_samr_Ids(struct ndr_print *ndr, const char *name, const struct samr_Ids *r);
void ndr_print_samr_GroupInfoAll(struct ndr_print *ndr, const char *name, const struct samr_GroupInfoAll *r);
void ndr_print_samr_GroupInfoAttributes(struct ndr_print *ndr, const char *name, const struct samr_GroupInfoAttributes *r);
void ndr_print_samr_GroupInfoEnum(struct ndr_print *ndr, const char *name, enum samr_GroupInfoEnum r);
void ndr_print_samr_GroupInfo(struct ndr_print *ndr, const char *name, const union samr_GroupInfo *r);
void ndr_print_samr_RidAttrArray(struct ndr_print *ndr, const char *name, const struct samr_RidAttrArray *r);
void ndr_print_samr_AliasInfoAll(struct ndr_print *ndr, const char *name, const struct samr_AliasInfoAll *r);
void ndr_print_samr_AliasInfoEnum(struct ndr_print *ndr, const char *name, enum samr_AliasInfoEnum r);
void ndr_print_samr_AliasInfo(struct ndr_print *ndr, const char *name, const union samr_AliasInfo *r);
void ndr_print_samr_UserInfoLevel(struct ndr_print *ndr, const char *name, enum samr_UserInfoLevel r);
void ndr_print_samr_UserInfo1(struct ndr_print *ndr, const char *name, const struct samr_UserInfo1 *r);
void ndr_print_samr_UserInfo2(struct ndr_print *ndr, const char *name, const struct samr_UserInfo2 *r);
enum ndr_err_code ndr_push_samr_LogonHours(struct ndr_push *ndr, ndr_flags_type ndr_flags, const struct samr_LogonHours *r);
enum ndr_err_code ndr_pull_samr_LogonHours(struct ndr_pull *ndr, ndr_flags_type ndr_flags, struct samr_LogonHours *r);
void ndr_print_samr_LogonHours(struct ndr_print *ndr, const char *name, const struct samr_LogonHours *r);
void ndr_print_samr_UserInfo3(struct ndr_print *ndr, const char *name, const struct samr_UserInfo3 *r);
void ndr_print_samr_UserInfo4(struct ndr_print *ndr, const char *name, const struct samr_UserInfo4 *r);
void ndr_print_samr_UserInfo5(struct ndr_print *ndr, const char *name, const struct samr_UserInfo5 *r);
void ndr_print_samr_UserInfo6(struct ndr_print *ndr, const char *name, const struct samr_UserInfo6 *r);
void ndr_print_samr_UserInfo7(struct ndr_print *ndr, const char *name, const struct samr_UserInfo7 *r);
void ndr_print_samr_UserInfo8(struct ndr_print *ndr, const char *name, const struct samr_UserInfo8 *r);
void ndr_print_samr_UserInfo9(struct ndr_print *ndr, const char *name, const struct samr_UserInfo9 *r);
void ndr_print_samr_UserInfo10(struct ndr_print *ndr, const char *name, const struct samr_UserInfo10 *r);
void ndr_print_samr_UserInfo11(struct ndr_print *ndr, const char *name, const struct samr_UserInfo11 *r);
void ndr_print_samr_UserInfo12(struct ndr_print *ndr, const char *name, const struct samr_UserInfo12 *r);
void ndr_print_samr_UserInfo13(struct ndr_print *ndr, const char *name, const struct samr_UserInfo13 *r);
void ndr_print_samr_UserInfo14(struct ndr_print *ndr, const char *name, const struct samr_UserInfo14 *r);
void ndr_print_samr_UserInfo16(struct ndr_print *ndr, const char *name, const struct samr_UserInfo16 *r);
void ndr_print_samr_UserInfo17(struct ndr_print *ndr, const char *name, const struct samr_UserInfo17 *r);
enum ndr_err_code ndr_push_samr_Password(struct ndr_push *ndr, ndr_flags_type ndr_flags, const struct samr_Password *r);
enum ndr_err_code ndr_pull_samr_Password(struct ndr_pull *ndr, ndr_flags_type ndr_flags, struct samr_Password *r);
void ndr_print_samr_Password(struct ndr_print *ndr, const char *name, const struct samr_Password *r);
void ndr_print_samr_UserInfo18(struct ndr_print *ndr, const char *name, const struct samr_UserInfo18 *r);
void ndr_print_samr_UserInfo20(struct ndr_print *ndr, const char *name, const struct samr_UserInfo20 *r);
void ndr_print_samr_FieldsPresent(struct ndr_print *ndr, const char *name, uint32_t r);
void ndr_print_samr_UserInfo21(struct ndr_print *ndr, const char *name, const struct samr_UserInfo21 *r);
enum ndr_err_code ndr_push_samr_CryptPassword(struct ndr_push *ndr, ndr_flags_type ndr_flags, const struct samr_CryptPassword *r);
enum ndr_err_code ndr_pull_samr_CryptPassword(struct ndr_pull *ndr, ndr_flags_type ndr_flags, struct samr_CryptPassword *r);
void ndr_print_samr_CryptPassword(struct ndr_print *ndr, const char *name, const struct samr_CryptPassword *r);
void ndr_print_samr_UserInfo23(struct ndr_print *ndr, const char *name, const struct samr_UserInfo23 *r);
void ndr_print_samr_UserInfo24(struct ndr_print *ndr, const char *name, const struct samr_UserInfo24 *r);
void ndr_print_samr_CryptPasswordEx(struct ndr_print *ndr, const char *name, const struct samr_CryptPasswordEx *r);
void ndr_print_samr_UserInfo25(struct ndr_print *ndr, const char *name, const struct samr_UserInfo25 *r);
void ndr_print_samr_UserInfo26(struct ndr_print *ndr, const char *name, const struct samr_UserInfo26 *r);
void ndr_print_samr_EncryptedPasswordAES(struct ndr_print *ndr, const char *name, const struct samr_EncryptedPasswordAES *r);
void ndr_print_samr_UserInfo31(struct ndr_print *ndr, const char *name, const struct samr_UserInfo31 *r);
void ndr_print_samr_UserInfo32(struct ndr_print *ndr, const char *name, const struct samr_UserInfo32 *r);
void ndr_print_samr_UserInfo(struct ndr_print *ndr, const char *name, const union samr_UserInfo *r);
enum ndr_err_code ndr_push_samr_RidWithAttribute(struct ndr_push *ndr, ndr_flags_type ndr_flags, const struct samr_RidWithAttribute *r);
enum ndr_err_code ndr_pull_samr_RidWithAttribute(struct ndr_pull *ndr, ndr_flags_type ndr_flags, struct samr_RidWithAttribute *r);
void ndr_print_samr_RidWithAttribute(struct ndr_print *ndr, const char *name, const struct samr_RidWithAttribute *r);
enum ndr_err_code ndr_push_samr_RidWithAttributeArray(struct ndr_push *ndr, ndr_flags_type ndr_flags, const struct samr_RidWithAttributeArray *r);
enum ndr_err_code ndr_pull_samr_RidWithAttributeArray(struct ndr_pull *ndr, ndr_flags_type ndr_flags, struct samr_RidWithAttributeArray *r);
void ndr_print_samr_RidWithAttributeArray(struct ndr_print *ndr, const char *name, const struct samr_RidWithAttributeArray *r);
void ndr_print_samr_DispEntryGeneral(struct ndr_print *ndr, const char *name, const struct samr_DispEntryGeneral *r);
void ndr_print_samr_DispInfoGeneral(struct ndr_print *ndr, const char *name, const struct samr_DispInfoGeneral *r);
void ndr_print_samr_DispEntryFull(struct ndr_print *ndr, const char *name, const struct samr_DispEntryFull *r);
void ndr_print_samr_DispInfoFull(struct ndr_print *ndr, const char *name, const struct samr_DispInfoFull *r);
void ndr_print_samr_DispEntryFullGroup(struct ndr_print *ndr, const char *name, const struct samr_DispEntryFullGroup *r);
void ndr_print_samr_DispInfoFullGroups(struct ndr_print *ndr, const char *name, const struct samr_DispInfoFullGroups *r);
void ndr_print_samr_DispEntryAscii(struct ndr_print *ndr, const char *name, const struct samr_DispEntryAscii *r);
void ndr_print_samr_DispInfoAscii(struct ndr_print *ndr, const char *name, const struct samr_DispInfoAscii *r);
void ndr_print_samr_DispInfo(struct ndr_print *ndr, const char *name, const union samr_DispInfo *r);
void ndr_print_samr_PwInfo(struct ndr_print *ndr, const char *name, const struct samr_PwInfo *r);
void ndr_print_samr_ConnectVersion(struct ndr_print *ndr, const char *name, enum samr_ConnectVersion r);
enum ndr_err_code ndr_push_samPwdChangeReason(struct ndr_push *ndr, ndr_flags_type ndr_flags, enum samPwdChangeReason r);
enum ndr_err_code ndr_pull_samPwdChangeReason(struct ndr_pull *ndr, ndr_flags_type ndr_flags, enum samPwdChangeReason *r);
void ndr_print_samPwdChangeReason(struct ndr_print *ndr, const char *name, enum samPwdChangeReason r);
void ndr_print_userPwdChangeFailureInformation(struct ndr_print *ndr, const char *name, const struct userPwdChangeFailureInformation *r);
void ndr_print_samr_SupportedFeatures(struct ndr_print *ndr, const char *name, enum samr_SupportedFeatures r);
void ndr_print_samr_ConnectInfo1(struct ndr_print *ndr, const char *name, const struct samr_ConnectInfo1 *r);
void ndr_print_samr_ConnectInfo(struct ndr_print *ndr, const char *name, const union samr_ConnectInfo *r);
void ndr_print_samr_ValidateFieldsPresent(struct ndr_print *ndr, const char *name, uint32_t r);
void ndr_print_samr_ValidatePasswordLevel(struct ndr_print *ndr, const char *name, enum samr_ValidatePasswordLevel r);
void ndr_print_samr_ValidationStatus(struct ndr_print *ndr, const char *name, enum samr_ValidationStatus r);
void ndr_print_samr_ValidationBlob(struct ndr_print *ndr, const char *name, const struct samr_ValidationBlob *r);
void ndr_print_samr_ValidatePasswordInfo(struct ndr_print *ndr, const char *name, const struct samr_ValidatePasswordInfo *r);
void ndr_print_samr_ValidatePasswordRepCtr(struct ndr_print *ndr, const char *name, const struct samr_ValidatePasswordRepCtr *r);
void ndr_print_samr_ValidatePasswordRep(struct ndr_print *ndr, const char *name, const union samr_ValidatePasswordRep *r);
void ndr_print_samr_ValidatePasswordReq3(struct ndr_print *ndr, const char *name, const struct samr_ValidatePasswordReq3 *r);
void ndr_print_samr_ValidatePasswordReq2(struct ndr_print *ndr, const char *name, const struct samr_ValidatePasswordReq2 *r);
void ndr_print_samr_ValidatePasswordReq1(struct ndr_print *ndr, const char *name, const struct samr_ValidatePasswordReq1 *r);
void ndr_print_samr_ValidatePasswordReq(struct ndr_print *ndr, const char *name, const union samr_ValidatePasswordReq *r);
void ndr_print_samr_Connect(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_Connect *r);
enum ndr_err_code ndr_push_samr_Close(struct ndr_push *ndr, ndr_flags_type flags, const struct samr_Close *r);
enum ndr_err_code ndr_pull_samr_Close(struct ndr_pull *ndr, ndr_flags_type flags, struct samr_Close *r);
void ndr_print_samr_Close(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_Close *r);
void ndr_print_samr_SetSecurity(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_SetSecurity *r);
void ndr_print_samr_QuerySecurity(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_QuerySecurity *r);
void ndr_print_samr_Shutdown(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_Shutdown *r);
void ndr_print_samr_LookupDomain(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_LookupDomain *r);
void ndr_print_samr_EnumDomains(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_EnumDomains *r);
enum ndr_err_code ndr_push_samr_OpenDomain(struct ndr_push *ndr, ndr_flags_type flags, const struct samr_OpenDomain *r);
enum ndr_err_code ndr_pull_samr_OpenDomain(struct ndr_pull *ndr, ndr_flags_type flags, struct samr_OpenDomain *r);
void ndr_print_samr_OpenDomain(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_OpenDomain *r);
void ndr_print_samr_QueryDomainInfo(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_QueryDomainInfo *r);
void ndr_print_samr_SetDomainInfo(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_SetDomainInfo *r);
void ndr_print_samr_CreateDomainGroup(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_CreateDomainGroup *r);
void ndr_print_samr_EnumDomainGroups(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_EnumDomainGroups *r);
void ndr_print_samr_CreateUser(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_CreateUser *r);
void ndr_print_samr_EnumDomainUsers(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_EnumDomainUsers *r);
void ndr_print_samr_CreateDomAlias(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_CreateDomAlias *r);
void ndr_print_samr_EnumDomainAliases(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_EnumDomainAliases *r);
void ndr_print_samr_GetAliasMembership(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_GetAliasMembership *r);
enum ndr_err_code ndr_push_samr_LookupNames(struct ndr_push *ndr, ndr_flags_type flags, const struct samr_LookupNames *r);
enum ndr_err_code ndr_pull_samr_LookupNames(struct ndr_pull *ndr, ndr_flags_type flags, struct samr_LookupNames *r);
void ndr_print_samr_LookupNames(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_LookupNames *r);
void ndr_print_samr_LookupRids(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_LookupRids *r);
void ndr_print_samr_OpenGroup(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_OpenGroup *r);
void ndr_print_samr_QueryGroupInfo(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_QueryGroupInfo *r);
void ndr_print_samr_SetGroupInfo(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_SetGroupInfo *r);
void ndr_print_samr_AddGroupMember(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_AddGroupMember *r);
void ndr_print_samr_DeleteDomainGroup(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_DeleteDomainGroup *r);
void ndr_print_samr_DeleteGroupMember(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_DeleteGroupMember *r);
void ndr_print_samr_QueryGroupMember(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_QueryGroupMember *r);
void ndr_print_samr_SetMemberAttributesOfGroup(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_SetMemberAttributesOfGroup *r);
void ndr_print_samr_OpenAlias(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_OpenAlias *r);
void ndr_print_samr_QueryAliasInfo(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_QueryAliasInfo *r);
void ndr_print_samr_SetAliasInfo(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_SetAliasInfo *r);
void ndr_print_samr_DeleteDomAlias(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_DeleteDomAlias *r);
void ndr_print_samr_AddAliasMember(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_AddAliasMember *r);
void ndr_print_samr_DeleteAliasMember(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_DeleteAliasMember *r);
void ndr_print_samr_GetMembersInAlias(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_GetMembersInAlias *r);
enum ndr_err_code ndr_push_samr_OpenUser(struct ndr_push *ndr, ndr_flags_type flags, const struct samr_OpenUser *r);
enum ndr_err_code ndr_pull_samr_OpenUser(struct ndr_pull *ndr, ndr_flags_type flags, struct samr_OpenUser *r);
void ndr_print_samr_OpenUser(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_OpenUser *r);
void ndr_print_samr_DeleteUser(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_DeleteUser *r);
enum ndr_err_code ndr_push_samr_QueryUserInfo(struct ndr_push *ndr, ndr_flags_type flags, const struct samr_QueryUserInfo *r);
enum ndr_err_code ndr_pull_samr_QueryUserInfo(struct ndr_pull *ndr, ndr_flags_type flags, struct samr_QueryUserInfo *r);
void ndr_print_samr_QueryUserInfo(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_QueryUserInfo *r);
enum ndr_err_code ndr_push_samr_SetUserInfo(struct ndr_push *ndr, ndr_flags_type flags, const struct samr_SetUserInfo *r);
enum ndr_err_code ndr_pull_samr_SetUserInfo(struct ndr_pull *ndr, ndr_flags_type flags, struct samr_SetUserInfo *r);
void ndr_print_samr_SetUserInfo(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_SetUserInfo *r);
void ndr_print_samr_ChangePasswordUser(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_ChangePasswordUser *r);
void ndr_print_samr_GetGroupsForUser(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_GetGroupsForUser *r);
void ndr_print_samr_QueryDisplayInfo(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_QueryDisplayInfo *r);
void ndr_print_samr_GetDisplayEnumerationIndex(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_GetDisplayEnumerationIndex *r);
void ndr_print_samr_TestPrivateFunctionsDomain(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_TestPrivateFunctionsDomain *r);
void ndr_print_samr_TestPrivateFunctionsUser(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_TestPrivateFunctionsUser *r);
enum ndr_err_code ndr_push_samr_GetUserPwInfo(struct ndr_push *ndr, ndr_flags_type flags, const struct samr_GetUserPwInfo *r);
enum ndr_err_code ndr_pull_samr_GetUserPwInfo(struct ndr_pull *ndr, ndr_flags_type flags, struct samr_GetUserPwInfo *r);
void ndr_print_samr_GetUserPwInfo(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_GetUserPwInfo *r);
void ndr_print_samr_RemoveMemberFromForeignDomain(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_RemoveMemberFromForeignDomain *r);
void ndr_print_samr_QueryDomainInfo2(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_QueryDomainInfo2 *r);
void ndr_print_samr_QueryUserInfo2(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_QueryUserInfo2 *r);
void ndr_print_samr_QueryDisplayInfo2(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_QueryDisplayInfo2 *r);
void ndr_print_samr_GetDisplayEnumerationIndex2(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_GetDisplayEnumerationIndex2 *r);
void ndr_print_samr_CreateUser2(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_CreateUser2 *r);
void ndr_print_samr_QueryDisplayInfo3(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_QueryDisplayInfo3 *r);
void ndr_print_samr_AddMultipleMembersToAlias(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_AddMultipleMembersToAlias *r);
void ndr_print_samr_RemoveMultipleMembersFromAlias(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_RemoveMultipleMembersFromAlias *r);
void ndr_print_samr_OemChangePasswordUser2(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_OemChangePasswordUser2 *r);
void ndr_print_samr_ChangePasswordUser2(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_ChangePasswordUser2 *r);
void ndr_print_samr_GetDomPwInfo(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_GetDomPwInfo *r);
void ndr_print_samr_Connect2(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_Connect2 *r);
enum ndr_err_code ndr_push_samr_SetUserInfo2(struct ndr_push *ndr, ndr_flags_type flags, const struct samr_SetUserInfo2 *r);
enum ndr_err_code ndr_pull_samr_SetUserInfo2(struct ndr_pull *ndr, ndr_flags_type flags, struct samr_SetUserInfo2 *r);
void ndr_print_samr_SetUserInfo2(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_SetUserInfo2 *r);
void ndr_print_samr_SetBootKeyInformation(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_SetBootKeyInformation *r);
void ndr_print_samr_GetBootKeyInformation(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_GetBootKeyInformation *r);
void ndr_print_samr_Connect3(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_Connect3 *r);
void ndr_print_samr_Connect4(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_Connect4 *r);
enum ndr_err_code ndr_push_samr_ChangePasswordUser3(struct ndr_push *ndr, ndr_flags_type flags, const struct samr_ChangePasswordUser3 *r);
enum ndr_err_code ndr_pull_samr_ChangePasswordUser3(struct ndr_pull *ndr, ndr_flags_type flags, struct samr_ChangePasswordUser3 *r);
void ndr_print_samr_ChangePasswordUser3(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_ChangePasswordUser3 *r);
enum ndr_err_code ndr_push_samr_Connect5(struct ndr_push *ndr, ndr_flags_type flags, const struct samr_Connect5 *r);
enum ndr_err_code ndr_pull_samr_Connect5(struct ndr_pull *ndr, ndr_flags_type flags, struct samr_Connect5 *r);
void ndr_print_samr_Connect5(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_Connect5 *r);
void ndr_print_samr_RidToSid(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_RidToSid *r);
void ndr_print_samr_SetDsrmPassword(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_SetDsrmPassword *r);
void ndr_print_samr_ValidatePassword(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_ValidatePassword *r);
void ndr_print_samr_Opnum68NotUsedOnWire(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_Opnum68NotUsedOnWire *r);
void ndr_print_samr_Opnum69NotUsedOnWire(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_Opnum69NotUsedOnWire *r);
void ndr_print_samr_Opnum70NotUsedOnWire(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_Opnum70NotUsedOnWire *r);
void ndr_print_samr_Opnum71NotUsedOnWire(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_Opnum71NotUsedOnWire *r);
void ndr_print_samr_Opnum72NotUsedOnWire(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_Opnum72NotUsedOnWire *r);
enum ndr_err_code ndr_push_samr_ChangePasswordUser4(struct ndr_push *ndr, ndr_flags_type flags, const struct samr_ChangePasswordUser4 *r);
enum ndr_err_code ndr_pull_samr_ChangePasswordUser4(struct ndr_pull *ndr, ndr_flags_type flags, struct samr_ChangePasswordUser4 *r);
void ndr_print_samr_ChangePasswordUser4(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct samr_ChangePasswordUser4 *r);
#endif /* _HEADER_NDR_samr */
