/* header auto-generated by pidl */

#include <ndr.h>
#include <gen_ndr/atsvc.h>

#ifndef _HEADER_NDR_atsvc
#define _HEADER_NDR_atsvc

#define NDR_ATSVC_UUID "1ff70682-0a51-30e8-076d-740be8cee98b"
#define NDR_ATSVC_VERSION 1.0
#define NDR_ATSVC_NAME "atsvc"
#define NDR_ATSVC_HELPSTRING "Microsoft AT-Scheduler Service"
extern const struct ndr_interface_table ndr_table_atsvc;
#define NDR_ATSVC_JOBADD (0x00)

#define NDR_ATSVC_JOBDEL (0x01)

#define NDR_ATSVC_JOBENUM (0x02)

#define NDR_ATSVC_JOBGETINFO (0x03)

#define NDR_ATSVC_CALL_COUNT (4)
void ndr_print_atsvc_DaysOfMonth(struct ndr_print *ndr, const char *name, uint32_t r);
void ndr_print_atsvc_Flags(struct ndr_print *ndr, const char *name, uint8_t r);
void ndr_print_atsvc_DaysOfWeek(struct ndr_print *ndr, const char *name, uint8_t r);
void ndr_print_atsvc_JobInfo(struct ndr_print *ndr, const char *name, const struct atsvc_JobInfo *r);
void ndr_print_atsvc_JobEnumInfo(struct ndr_print *ndr, const char *name, const struct atsvc_JobEnumInfo *r);
void ndr_print_atsvc_enum_ctr(struct ndr_print *ndr, const char *name, const struct atsvc_enum_ctr *r);
enum ndr_err_code ndr_push_atsvc_JobAdd(struct ndr_push *ndr, ndr_flags_type flags, const struct atsvc_JobAdd *r);
enum ndr_err_code ndr_pull_atsvc_JobAdd(struct ndr_pull *ndr, ndr_flags_type flags, struct atsvc_JobAdd *r);
void ndr_print_atsvc_JobAdd(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct atsvc_JobAdd *r);
enum ndr_err_code ndr_push_atsvc_JobDel(struct ndr_push *ndr, ndr_flags_type flags, const struct atsvc_JobDel *r);
enum ndr_err_code ndr_pull_atsvc_JobDel(struct ndr_pull *ndr, ndr_flags_type flags, struct atsvc_JobDel *r);
void ndr_print_atsvc_JobDel(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct atsvc_JobDel *r);
enum ndr_err_code ndr_push_atsvc_JobEnum(struct ndr_push *ndr, ndr_flags_type flags, const struct atsvc_JobEnum *r);
enum ndr_err_code ndr_pull_atsvc_JobEnum(struct ndr_pull *ndr, ndr_flags_type flags, struct atsvc_JobEnum *r);
void ndr_print_atsvc_JobEnum(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct atsvc_JobEnum *r);
enum ndr_err_code ndr_push_atsvc_JobGetInfo(struct ndr_push *ndr, ndr_flags_type flags, const struct atsvc_JobGetInfo *r);
enum ndr_err_code ndr_pull_atsvc_JobGetInfo(struct ndr_pull *ndr, ndr_flags_type flags, struct atsvc_JobGetInfo *r);
void ndr_print_atsvc_JobGetInfo(struct ndr_print *ndr, const char *name, ndr_flags_type flags, const struct atsvc_JobGetInfo *r);
#endif /* _HEADER_NDR_atsvc */
